% ViPrice Toolbox 0.3
% Copyright (C) 2015  Peer Neubert, peer.neubert@etit.tu-chemnitz.de
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
classdef StarHoughClass  < handle
    % A call of compareStarHough(L1, L2) updates the member variables
    
    properties
        H % the resulting Hough space
        D % the pairwise landmark distance
        S % the pairwise landmark similarity
        star1 % star model generated from first landmark set in compareStarHough()
        star2 % star model generated from second landmark set in compareStarHough()
        idx2_set % idx2_set{idx1} is the vector of indices ni L2 of 
                 % top-k-left-right matchings of landmark idx1 from L1
        params % the parameters         
    end    
    
    methods
        
        function obj = StarHoughClass(obj)
            obj.params.radiusHough = 20;
            obj.params.minLandmarkSim = 0;
            obj.params.smoothStd = max(5,obj.params.radiusHough/4);
            obj.params.smoothSize = obj.params.radiusHough/10;
            
        end
        
        % compute similarity of star models from their parts distances and part
        % midpoint coordinates       
        %              
        function [starSim] = compareStarHough(obj, L1, L2, D)

          if ~exist('D', 'var')
            obj.D = pdist2(L1.D, L2.D, 'cosine');
          else
            obj.D = D;  
          end            
            
          % select potential matchings using left-right check
          obj.idx2_set = obj.getTopKLeftRightChecked(obj.D,1);

          % compute the star model
          % this is simply a vector [x,y] for each feature marking the relative
          % position of this feature to the midpoint (mean) of all features
          [midX1, midY1] = obj.getMidpoints(L1);
          [midX2, midY2] = obj.getMidpoints(L2);
          obj.star1 = obj.getStarModel(midX1, midY1);
          obj.star2 = obj.getStarModel(midX2, midY2);

          % distribute probability mass according to descriptor similarity
          obj.H = zeros(2*obj.params.radiusHough+1);

          % get similarity from distance D
          obj.S = 2-obj.D; 

          % iterate over D and update Hough space
          for idx1=1:size(obj.D,1)
            sumAllSims = sum( obj.S(idx1,:) );

            % use potential matchings from prior left-right check
            for idx2=obj.idx2_set{idx1}
            
              % descriptor similarity
              pairSim = obj.S(idx1,idx2);
            
              % apply additional threshold on similarity
              if pairSim<obj.params.minLandmarkSim
               continue;
              end

              sim = pairSim * pairSim/ sumAllSims;

              % compute update 
              U = zeros(size(obj.H));
              xH = round(obj.getHoughCO(obj.star1(idx1,1), obj.star2(idx2,1), obj.params.radiusHough));
              yH = round(obj.getHoughCO(obj.star1(idx1,2), obj.star2(idx2,2), obj.params.radiusHough));
        %       U(yH, xH) = sim; % pose prior: * (1-norm( star1(idx1,:) - star2(idx2,:)));   
              U(yH, xH) = sim;

              % update H
              obj.H = obj.H + U;

            end
          end

          obj.H = imfilter(obj.H, fspecial('gaussian', obj.params.smoothSize, obj.params.smoothStd));

          % return maximum value of H
          starSim = max(obj.H(:));
        end

        % Return the best matchings based on the internal results of the last 
        % call of compareStarHough()
        %
        % M ... each row: [idx1 idx2 score pairSim houghVal], index in L1, 
        %       L2 given to last call of compareStarHough()
        function M = recallBestMatchings(obj)
            
            % evaluate accordance with Houg accumulator result for each 
            % left-right matching 
            M = [];
            for idx1=1:size(obj.D,1)
                sumAllSims = sum( obj.S(idx1,:) );

                % use potential matchings from prior left-right check
                for idx2=obj.idx2_set{idx1}

                  % descriptor similarity
                  pairSim = obj.S(idx1,idx2);
                  
                  % Hough value
                  xH = round(obj.getHoughCO(obj.star1(idx1,1), obj.star2(idx2,1), obj.params.radiusHough));
                  yH = round(obj.getHoughCO(obj.star1(idx1,2), obj.star2(idx2,2), obj.params.radiusHough));
                  houghVal = obj.H(yH, xH);
                  
                  M(end+1, :) = [idx1 idx2 pairSim*houghVal pairSim houghVal];
                  
                end
            end 
              
            [~, idx] = sort(M(:,3), 'descend');
            
            M = M(idx, :);
        end
        
        % Create a visualization of the best matchings
        % k .. maximum number of matchings
        % cm ... colormap
        function V = showBestMatchings(obj, I1, I2, L1, L2, k, cm)
            M = obj.recallBestMatchings();
            
            h_offset = size(I1, 2);
            
            scale = size(I1,1) / size(I2,1);
            V = cat(2, I1, imresize(I2, scale));
            
            linewidth = 10;
            
            if ~exist('cm', 'var')
                cm = 255*colormap(lines);
            end
            
            for i=1:min(k, size(M,1))
                color = cm(i,: );
                idx1 = M(i,1);                
                idx2 = M(i,2);
                
                % show landmarks
                M1 = RegionVisualizerClass.getMaskBoundary(L1.regions{idx1}.mask, 2*linewidth);
                V = RegionVisualizerClass.drawMaskAndBBOnImageRoi(V, M1, color, 0.5, L1.regions{idx1}.ROI, linewidth);
                
                ROI2 = (L2.regions{idx2}.ROI * scale) + [h_offset 0 h_offset 0];   
                mask2 = imresize(L2.regions{idx2}.mask, [ROI2(4)-ROI2(2)+1 ROI2(3)-ROI2(1)+1] );
                
                M2 = RegionVisualizerClass.getMaskBoundary(mask2, 2*linewidth);
                V = RegionVisualizerClass.drawMaskAndBBOnImageRoi(V, M2, color, 0.5, ROI2, linewidth);                
                
                % show matching
                                
            end
            
        end
        
        % get for idx1 a list of idx2 such that idx1 is amongst the top-k-matched
        % of idx2 and vice versa
        function idx2_set = getTopKLeftRightChecked(obj, D, k)
            
          % find k-best matches for idx1 --> each row are indices
          [~, S1] = sort(D,2);

          % find k-best matches for idx2 --> each row are indices
          [~, S2] = sort(D',2);

          % For each idx1 have a look at the best k matchings potential_idx2.
          % if idx1 is in the first k matchings of potential_idx2, then add
          % potential_idx2 to idx2_set{idx1}
          idx2_set = {};
          for idx1=1:size(D,1)
            k_cur = min(k, size(D,2));      
            potential_idx2_set = S1(idx1, 1:k_cur);
            selected_idx2_vec = [];
            for potential_idx2 = potential_idx2_set
              if idx1 == S2(potential_idx2, 1:k_cur)
                selected_idx2_vec = [selected_idx2_vec potential_idx2];
              end
            end
            idx2_set{idx1} = selected_idx2_vec;    
          end

        end
        
        % compute the coordinate in Hough space 
        %
        % the center H(radiusHough+1, radiusHough+1) of H is the location of the
        % target star model star2
        % In worst case we can map a very right 0.5 part of star1 to a very left -0.5
        % part of star2. Thus, the radiusHough spans the complete model
        % to map [-0.5, 0.5] star model space
        function h = getHoughCO(obj, x1,x2,r)
          % in principle
          % x1Hough = xStar1*radiusHough;
          % x2Hough = xStar2*radiusHough + x1Hough;  
          % and additional shift to Hough image   
          h = r + 1 + (x2-x1)*r;  
        end

        function plotStar(star)  
          for i=1:size(star,1)
            axis ij;
            plot([star(i,1), 0], [star(i,2), 0],'b-');
            hold on;
            text(star(i,1), star(i,2), num2str(i), 'fontsize', 20);   

          end
        end
        
        % Create two column vectors of coordinates of all region midpoints 
        function [midX, midY] = getMidpoints(obj, L)
            midX = zeros(numel(L.regions),1);
            midY = zeros(numel(L.regions),1);
            for i=1:numel(L.regions)
                midX(i) = L.regions{i}.midpoint(1);
                midY(i) = L.regions{i}.midpoint(2);
            end                
        end
        
        % compute the star model
        % this is simply a vector [x,y] for each feature marking the relative
        % position of this feature to the midpoint (mean) of all features
        % midX, midY ... column vectors of mid point coordinates
        % star ... each row [x,y] in range [-0.5 0.5]
        function star = getStarModel(obj, midX, midY)

          if size(midX,2)>1
            fprintf('getStarModel(): midX should be column vector\n');
          end

          minX = min(midX);
          maxX = max(midX);
          minY = min(midY);
          maxY = max(midY);

          % if we realy use mean, we can not gurantee range [-0.5, 0.5]
          meanX = (maxX + minX) / 2; %mean(midX);
          meanY = (maxY + minY) / 2; %mean(midY);

          % compute X and Y star coordinates with check of zero width
          if abs(maxX-minX) > 0 
            starX = (midX-meanX)./(maxX-minX);
          else
            starX = zeros(size(midX));
          end

          if abs(maxY-minY) > 0 
            starY = (midY-meanY)./(maxY-minY);
          else
            starY = zeros(size(midY));
          end

          star = [ starX starY ];

        end
    end
end
